@@ -31,6 +31,7 @@
 #define CB_MSGBUFSIZ   256               /* size of a buffer for log message */
 #define CB_IOBUFSIZ    8192              /* size of an I/O buffer */
 #define CB_FILEMODE    00644             /* permission of a creating file */
+#define CB_NUMBUFSIZ   32                /* size of a buffer for a number */
 #define CB_ENCBUFSIZ   32                /* size of a buffer for encoding name */
 #define CB_DATEBUFSIZ  64                /* size of a buffer for date expression */
 #define CB_VNUMBUFSIZ  8                 /* size of a buffer for variable length number */
@@ -729,7 +730,7 @@ void cblistover(CBLIST *list, int index, const char *ptr, int size){
 /* Sort elements of a list in lexical order. */
 void cblistsort(CBLIST *list){
   assert(list);
-  cbqsort(list->array + list->start, list->num, sizeof(list->array[0]), cblistelemcmp);
+  qsort(list->array + list->start, list->num, sizeof(list->array[0]), cblistelemcmp);
 }
 
 
@@ -838,7 +839,7 @@ void cbmapclose(CBMAP *map){
   CBMAPDATUM *datum, *next;
   datum = map->first;
   while(datum){
-    next = (CBMAPDATUM *)(datum->next);
+    next = datum->next;
     free(datum->kbuf);
     free(datum->vbuf);
     free(datum);
@@ -862,19 +863,19 @@ int cbmapput(CBMAP *map, const char *kbuf, int ksiz, const char *vbuf, int vsiz,
   hash = cbsecondhash(kbuf, ksiz);
   while(datum){
     if(hash > datum->hash){
-      entp = (CBMAPDATUM **)&(datum->left);
-      datum = (CBMAPDATUM *)datum->left;
+      entp = &(datum->left);
+      datum = datum->left;
     } else if(hash < datum->hash){
-      entp = (CBMAPDATUM **)&(datum->right);
-      datum = (CBMAPDATUM *)datum->right;
+      entp = &(datum->right);
+      datum = datum->right;
     } else {
       kcmp = cbkeycmp(kbuf, ksiz, datum->kbuf, datum->ksiz);
       if(kcmp < 0){
-        entp = (CBMAPDATUM **)&(datum->left);
-        datum = (CBMAPDATUM *)datum->left;
+        entp = &(datum->left);
+        datum = datum->left;
       } else if(kcmp > 0){
-        entp = (CBMAPDATUM **)&(datum->right);
-        datum = (CBMAPDATUM *)datum->right;
+        entp = &(datum->right);
+        datum = datum->right;
       } else {
         if(!over) return FALSE;
         if(vsiz > datum->vsiz){
@@ -897,11 +898,11 @@ int cbmapput(CBMAP *map, const char *kbuf, int ksiz, const char *vbuf, int vsiz,
   datum->hash = hash;
   datum->left = NULL;
   datum->right = NULL;
-  datum->prev = (char *)map->last;
+  datum->prev = map->last;
   datum->next = NULL;
   *entp = datum;
   if(!map->first) map->first = datum;
-  if(map->last) map->last->next = (char *)datum;
+  if(map->last) map->last->next = datum;
   map->last = datum;
   map->rnum++;
   return TRUE;
@@ -921,19 +922,19 @@ void cbmapputcat(CBMAP *map, const char *kbuf, int ksiz, const char *vbuf, int v
   hash = cbsecondhash(kbuf, ksiz);
   while(datum){
     if(hash > datum->hash){
-      entp = (CBMAPDATUM **)&(datum->left);
-      datum = (CBMAPDATUM *)datum->left;
+      entp = &(datum->left);
+      datum = datum->left;
     } else if(hash < datum->hash){
-      entp = (CBMAPDATUM **)&(datum->right);
-      datum = (CBMAPDATUM *)datum->right;
+      entp = &(datum->right);
+      datum = datum->right;
     } else {
       kcmp = cbkeycmp(kbuf, ksiz, datum->kbuf, datum->ksiz);
       if(kcmp < 0){
-        entp = (CBMAPDATUM **)&(datum->left);
-        datum = (CBMAPDATUM *)datum->left;
+        entp = &(datum->left);
+        datum = datum->left;
       } else if(kcmp > 0){
-        entp = (CBMAPDATUM **)&(datum->right);
-        datum = (CBMAPDATUM *)datum->right;
+        entp = &(datum->right);
+        datum = datum->right;
       } else {
         asiz = datum->vsiz + vsiz;
         unit = asiz <= CB_MAPCSUNIT ? CB_MAPCSUNIT : CB_MAPCBUNIT;
@@ -958,11 +959,11 @@ void cbmapputcat(CBMAP *map, const char *kbuf, int ksiz, const char *vbuf, int v
   datum->hash = hash;
   datum->left = NULL;
   datum->right = NULL;
-  datum->prev = (char *)map->last;
+  datum->prev = map->last;
   datum->next = NULL;
   *entp = datum;
   if(!map->first) map->first = datum;
-  if(map->last) map->last->next = (char *)datum;
+  if(map->last) map->last->next = datum;
   map->last = datum;
   map->rnum++;
 }
@@ -980,44 +981,44 @@ int cbmapout(CBMAP *map, const char *kbuf, int ksiz){
   hash = cbsecondhash(kbuf, ksiz);
   while(datum){
     if(hash > datum->hash){
-      entp = (CBMAPDATUM **)&(datum->left);
-      datum = (CBMAPDATUM *)datum->left;
+      entp = &(datum->left);
+      datum = datum->left;
     } else if(hash < datum->hash){
-      entp = (CBMAPDATUM **)&(datum->right);
-      datum = (CBMAPDATUM *)datum->right;
+      entp = &(datum->right);
+      datum = datum->right;
     } else {
       kcmp = cbkeycmp(kbuf, ksiz, datum->kbuf, datum->ksiz);
       if(kcmp < 0){
-        entp = (CBMAPDATUM **)&(datum->left);
-        datum = (CBMAPDATUM *)datum->left;
+        entp = &(datum->left);
+        datum = datum->left;
       } else if(kcmp > 0){
-        entp = (CBMAPDATUM **)&(datum->right);
-        datum = (CBMAPDATUM *)datum->right;
+        entp = &(datum->right);
+        datum = datum->right;
       } else {
-        if(datum->prev) ((CBMAPDATUM *)(datum->prev))->next = datum->next;
-        if(datum->next) ((CBMAPDATUM *)(datum->next))->prev = datum->prev;
-        if(datum == map->first) map->first = (CBMAPDATUM *)datum->next;
-        if(datum == map->last) map->last = (CBMAPDATUM *)datum->prev;
+        if(datum->prev) datum->prev->next = datum->next;
+        if(datum->next) datum->next->prev = datum->prev;
+        if(datum == map->first) map->first = datum->next;
+        if(datum == map->last) map->last = datum->prev;
         if(datum->left && !datum->right){
-          *entp = (CBMAPDATUM *)datum->left;
+          *entp = datum->left;
         } else if(!datum->left && datum->right){
-          *entp = (CBMAPDATUM *)datum->right;
+          *entp = datum->right;
         } else if(!datum->left && !datum->left){
           *entp = NULL;
         } else {
-          *entp = (CBMAPDATUM *)datum->left;
+          *entp = datum->left;
           tmp = *entp;
           while(TRUE){
             if(hash > tmp->hash){
               if(tmp->left){
-                tmp = (CBMAPDATUM *)tmp->left;
+                tmp = tmp->left;
               } else {
                 tmp->left = datum->right;
                 break;
               }
             } else if(hash < tmp->hash){
               if(tmp->right){
-                tmp = (CBMAPDATUM *)tmp->right;
+                tmp = tmp->right;
               } else {
                 tmp->right = datum->right;
                 break;
@@ -1026,14 +1027,14 @@ int cbmapout(CBMAP *map, const char *kbuf, int ksiz){
               kcmp = cbkeycmp(kbuf, ksiz, datum->kbuf, datum->ksiz);
               if(kcmp < 0){
                 if(tmp->left){
-                  tmp = (CBMAPDATUM *)tmp->left;
+                  tmp = tmp->left;
                 } else {
                   tmp->left = datum->right;
                   break;
                 }
               } else {
                 if(tmp->right){
-                  tmp = (CBMAPDATUM *)tmp->right;
+                  tmp = tmp->right;
                 } else {
                   tmp->right = datum->right;
                   break;
@@ -1064,15 +1065,15 @@ const char *cbmapget(const CBMAP *map, const char *kbuf, int ksiz, int *sp){
   hash = cbsecondhash(kbuf, ksiz);
   while(datum){
     if(hash > datum->hash){
-      datum = (CBMAPDATUM *)datum->left;
+      datum = datum->left;
     } else if(hash < datum->hash){
-      datum = (CBMAPDATUM *)datum->right;
+      datum = datum->right;
     } else {
       kcmp = cbkeycmp(kbuf, ksiz, datum->kbuf, datum->ksiz);
       if(kcmp < 0){
-        datum = (CBMAPDATUM *)datum->left;
+        datum = datum->left;
       } else if(kcmp > 0){
-        datum = (CBMAPDATUM *)datum->right;
+        datum = datum->right;
       } else {
         if(sp) *sp = datum->vsiz;
         return datum->vbuf;
@@ -1093,33 +1094,33 @@ int cbmapmove(CBMAP *map, const char *kbuf, int ksiz, int head){
   hash = cbsecondhash(kbuf, ksiz);
   while(datum){
     if(hash > datum->hash){
-      datum = (CBMAPDATUM *)datum->left;
+      datum = datum->left;
     } else if(hash < datum->hash){
-      datum = (CBMAPDATUM *)datum->right;
+      datum = datum->right;
     } else {
       kcmp = cbkeycmp(kbuf, ksiz, datum->kbuf, datum->ksiz);
       if(kcmp < 0){
-        datum = (CBMAPDATUM *)datum->left;
+        datum = datum->left;
       } else if(kcmp > 0){
-        datum = (CBMAPDATUM *)datum->right;
+        datum = datum->right;
       } else {
         if(head){
           if(map->first == datum) return TRUE;
-          if(map->last == datum) map->last = (CBMAPDATUM *)(datum->prev);
-          if(datum->prev) ((CBMAPDATUM *)(datum->prev))->next = datum->next;
-          if(datum->next) ((CBMAPDATUM *)(datum->next))->prev = datum->prev;
+          if(map->last == datum) map->last = datum->prev;
+          if(datum->prev) datum->prev->next = datum->next;
+          if(datum->next) datum->next->prev = datum->prev;
           datum->prev = NULL;
-          datum->next = (char *)(map->first);
-          map->first->prev = (char *)datum;
+          datum->next = map->first;
+          map->first->prev = datum;
           map->first = datum;
         } else {
           if(map->last == datum) return TRUE;
-          if(map->first == datum) map->first = (CBMAPDATUM *)(datum->next);
-          if(datum->prev) ((CBMAPDATUM *)(datum->prev))->next = datum->next;
-          if(datum->next) ((CBMAPDATUM *)(datum->next))->prev = datum->prev;
-          datum->prev = (char *)(map->last);
+          if(map->first == datum) map->first = datum->next;
+          if(datum->prev) datum->prev->next = datum->next;
+          if(datum->next) datum->next->prev = datum->prev;
+          datum->prev = map->last;
           datum->next = NULL;
-          map->last->next = (char *)datum;
+          map->last->next = datum;
           map->last = datum;
         }
         return TRUE;
@@ -1143,7 +1144,7 @@ const char *cbmapiternext(CBMAP *map, int *sp){
   assert(map);
   if(!map->cur) return NULL;
   datum = map->cur;
-  map->cur = (CBMAPDATUM *)datum->next;
+  map->cur = datum->next;
   if(sp) *sp = datum->ksiz;
   return datum->kbuf;
 }
@@ -1535,37 +1536,56 @@ CBLIST *cbdirlist(const char *name){
 
 
 /* Get the status of a file or a directory. */
-int cbfilestat(const char *name, int *isdirp, int *sizep, int *mtimep){
+int cbfilestat(const char *name, int *isdirp, int *sizep, time_t *mtimep){
   struct stat sbuf;
   assert(name);
   if(lstat(name, &sbuf) == -1) return FALSE;
   if(isdirp) *isdirp = S_ISDIR(sbuf.st_mode);
   if(sizep) *sizep = (int)sbuf.st_size;
-  if(mtimep) *mtimep = (int)sbuf.st_mtime;
+  if(mtimep) *mtimep = sbuf.st_mtime;
   return TRUE;
 }
 
 
+/* Remove a file or a directory and its sub ones recursively. */
+int cbremove(const char *name){
+  CBLIST *list;
+  const char *elem;
+  char *path;
+  int i, err, tail;
+  struct stat sbuf;
+  assert(name);
+  if(lstat(name, &sbuf) == -1) return FALSE;
+  if(unlink(name) == 0) return TRUE;
+  if(!S_ISDIR(sbuf.st_mode) || !(list = cbdirlist(name))) return FALSE;
+  err = FALSE;
+  tail = name[0] != '\0' && name[strlen(name)-1] == MYPATHCHR;
+  for(i = 0; i < CB_LISTNUM(list); i++){
+    elem = CB_LISTVAL(list, i, NULL);
+    if(!strcmp(MYCDIRSTR, elem) || !strcmp(MYPDIRSTR, elem)) continue;
+    if(tail){
+      path = cbsprintf("%s%s", name, elem);
+    } else {
+      path = cbsprintf("%s%c%s", name, MYPATHCHR, elem);
+    }
+    if(!cbremove(path)) err = TRUE;
+    free(path);
+  }
+  cblistclose(list);
+  return rmdir(name) == 0 ? TRUE : FALSE;
+}
+
+
 /* Break up a URL into elements. */
 CBMAP *cburlbreak(const char *str){
   CBMAP *map;
   char *tmp, *ep;
   const char *rp;
-  int i, serv;
+  int serv;
   assert(str);
   map = cbmapopenex(CB_MAPPBNUM);
-  rp = str;
-  while(*rp != '\0' && strchr(" \t\r\n\v\f", *rp)){
-    rp++;
-  }
-  tmp = cbmemdup(rp, -1);
-  for(i = 0; tmp[i] != '\0'; i++){
-    if(strchr(" \t\r\n\v\f", tmp[i])){
-      tmp[i] = '\0';
-      break;
-    }
-  }
-  rp = tmp;
+  tmp = cbmemdup(str, -1);
+  rp = cbstrtrim(tmp);
   cbmapput(map, "self", -1, rp, -1, TRUE);
   serv = FALSE;
   if(cbstrfwimatch(rp, "http://")){
@@ -1580,9 +1600,30 @@ CBMAP *cburlbreak(const char *str){
     cbmapput(map, "scheme", -1, "ftp", -1, TRUE);
     rp += 6;
     serv = TRUE;
+  } else if(cbstrfwimatch(rp, "sftp://")){
+    cbmapput(map, "scheme", -1, "sftp", -1, TRUE);
+    rp += 7;
+    serv = TRUE;
+  } else if(cbstrfwimatch(rp, "ftps://")){
+    cbmapput(map, "scheme", -1, "ftps", -1, TRUE);
+    rp += 7;
+    serv = TRUE;
+  } else if(cbstrfwimatch(rp, "tftp://")){
+    cbmapput(map, "scheme", -1, "tftp", -1, TRUE);
+    rp += 7;
+    serv = TRUE;
+  } else if(cbstrfwimatch(rp, "ldap://")){
+    cbmapput(map, "scheme", -1, "ldap", -1, TRUE);
+    rp += 7;
+    serv = TRUE;
+  } else if(cbstrfwimatch(rp, "ldaps://")){
+    cbmapput(map, "scheme", -1, "ldaps", -1, TRUE);
+    rp += 8;
+    serv = TRUE;
   } else if(cbstrfwimatch(rp, "file://")){
     cbmapput(map, "scheme", -1, "file", -1, TRUE);
     rp += 7;
+    serv = TRUE;
   }
   if((ep = strchr(rp, '#')) != NULL){
     cbmapput(map, "fragment", -1, ep + 1, -1, TRUE);
@@ -1629,121 +1670,164 @@ CBMAP *cburlbreak(const char *str){
 /* Resolve a relative URL with another absolute URL. */
 char *cburlresolve(const char *base, const char *target){
   CBMAP *telems, *belems;
-  CBLIST *bpaths, *opaths;
+  CBLIST *bpaths, *opaths, *qelems;
   CBDATUM *rbuf;
   const char *vbuf, *path;
-  char *tmp;
-  int i, vsiz;
+  char *tmp, *wp, *enc, numbuf[CB_NUMBUFSIZ];
+  int i, vsiz, port, num;
   assert(base && target);
+  while(*base > '\0' && *base <= ' '){
+    base++;
+  }
+  while(*target > '\0' && *target <= ' '){
+    target++;
+  }
+  if(*target == '\0') target = base;
   rbuf = cbdatumopen("", 0);
   telems = cburlbreak(target);
-  if((vbuf = cbmapget(telems, "scheme", -1, &vsiz)) != NULL){
+  port = 80;
+  belems = cburlbreak(cbmapget(telems, "scheme", -1, &vsiz) ? target : base);
+  if((vbuf = cbmapget(belems, "scheme", -1, &vsiz)) != NULL){
     cbdatumcat(rbuf, vbuf, vsiz);
     cbdatumcat(rbuf, "://", -1);
-    if((vbuf = cbmapget(telems, "authority", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, vbuf, vsiz);
-      cbdatumcat(rbuf, "@", 1);
-    }
-    if((vbuf = cbmapget(telems, "host", -1, &vsiz)) != NULL){
-      tmp = cbmemdup(vbuf, vsiz);
-      cbstrtolower(tmp);
-      cbdatumcat(rbuf, tmp, -1);
-      free(tmp);
-    } else {
-      cbdatumcat(rbuf, "localhost", -1);
-    }
-    if((vbuf = cbmapget(telems, "port", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, ":", 1);
-      cbdatumcat(rbuf, vbuf, vsiz);
-    }
-    if((vbuf = cbmapget(telems, "path", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, vbuf, vsiz);
-    } else {
-      cbdatumcat(rbuf, "/", 1);
-    }
-    if((vbuf = cbmapget(telems, "query", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, "?", 1);
-      cbdatumcat(rbuf, vbuf, vsiz);
-    }
-    if((vbuf = cbmapget(telems, "fragment", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, "#", 1);
-      cbdatumcat(rbuf, vbuf, vsiz);
+    if(!cbstricmp(vbuf, "https")){
+      port = 443;
+    } else if(!cbstricmp(vbuf, "ftp")){
+      port = 21;
+    } else if(!cbstricmp(vbuf, "sftp")){
+      port = 115;
+    } else if(!cbstricmp(vbuf, "ftps")){
+      port = 22;
+    } else if(!cbstricmp(vbuf, "tftp")){
+      port = 69;
+    } else if(!cbstricmp(vbuf, "ldap")){
+      port = 389;
+    } else if(!cbstricmp(vbuf, "ldaps")){
+      port = 636;
     }
   } else {
-    belems = cburlbreak(base);
-    if((vbuf = cbmapget(belems, "scheme", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, vbuf, vsiz);
-      cbdatumcat(rbuf, "://", -1);
-    } else {
-      cbdatumcat(rbuf, "http://", -1);
-    }
-    if((vbuf = cbmapget(belems, "authority", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, vbuf, vsiz);
-      cbdatumcat(rbuf, "@", 1);
-    }
-    if((vbuf = cbmapget(belems, "host", -1, &vsiz)) != NULL){
-      tmp = cbmemdup(vbuf, vsiz);
-      cbstrtolower(tmp);
-      cbdatumcat(rbuf, tmp, -1);
+    cbdatumcat(rbuf, "http://", -1);
+  }
+  if((vbuf = cbmapget(belems, "authority", -1, &vsiz)) != NULL){
+    if((wp = strchr(vbuf, ':')) != NULL){
+      *wp = '\0';
+      tmp = cburldecode(vbuf, NULL);
+      enc = cburlencode(tmp, -1);
+      cbdatumcat(rbuf, enc, -1);
+      free(enc);
+      free(tmp);
+      cbdatumcat(rbuf, ":", 1);
+      wp++;
+      tmp = cburldecode(wp, NULL);
+      enc = cburlencode(tmp, -1);
+      cbdatumcat(rbuf, enc, -1);
+      free(enc);
       free(tmp);
     } else {
-      cbdatumcat(rbuf, "localhost", -1);
+      tmp = cburldecode(vbuf, NULL);
+      enc = cburlencode(tmp, -1);
+      cbdatumcat(rbuf, enc, -1);
+      free(enc);
+      free(tmp);
     }
-    if((vbuf = cbmapget(belems, "port", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, ":", 1);
-      cbdatumcat(rbuf, vbuf, vsiz);
-    }
-    if(!(path = cbmapget(telems, "path", -1, NULL))) path = "/";
-    if(path[0] == '\0' && (vbuf = cbmapget(belems, "path", -1, &vsiz)) != NULL) path = vbuf;
-    if(path[0] == '/'){
-      cbdatumcat(rbuf, path, -1);
-    } else {
-      bpaths = cblistopen();
-      if((vbuf = cbmapget(belems, "path", -1, &vsiz)) != NULL){
-        opaths = cbsplit(vbuf, vsiz, "/");
-      } else {
-        opaths = cbsplit("/", 1, "/");
-      }
-      free(cblistpop(opaths, NULL));
-      for(i = 0; i < CB_LISTNUM(opaths); i++){
-        vbuf = CB_LISTVAL2(opaths, i, &vsiz);
-        if(vsiz < 1 || !strcmp(vbuf, ".")) continue;
-        if(!strcmp(vbuf, "..")){
-          free(cblistpop(bpaths, NULL));
-        } else {
-          cblistpush(bpaths, vbuf, vsiz);
-        }
-      }
-      cblistclose(opaths);
-      opaths = cbsplit(path, -1, "/");
-      for(i = 0; i < CB_LISTNUM(opaths); i++){
-        vbuf = CB_LISTVAL2(opaths, i, &vsiz);
-        if(vsiz < 1 || !strcmp(vbuf, ".")) continue;
-        if(!strcmp(vbuf, "..")){
-          free(cblistpop(bpaths, NULL));
-        } else {
-          cblistpush(bpaths, vbuf, vsiz);
-        }
-      }
-      cblistclose(opaths);
-      for(i = 0; i < CB_LISTNUM(bpaths); i++){
-        vbuf = CB_LISTVAL2(bpaths, i, &vsiz);
-        cbdatumcat(rbuf, "/", 1);
-        cbdatumcat(rbuf, vbuf, vsiz);
-      }
-      if(!(vbuf = cbmapget(telems, "file", -1, &vsiz)) || vsiz < 1) cbdatumcat(rbuf, "/", 1);
-      cblistclose(bpaths);
-    }
-    if((vbuf = cbmapget(telems, "query", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, "?", 1);
-      cbdatumcat(rbuf, vbuf, vsiz);
-    }
-    if((vbuf = cbmapget(telems, "fragment", -1, &vsiz)) != NULL){
-      cbdatumcat(rbuf, "#", 1);
-      cbdatumcat(rbuf, vbuf, vsiz);
-    }
-    cbmapclose(belems);
+    cbdatumcat(rbuf, "@", 1);
   }
+  if((vbuf = cbmapget(belems, "host", -1, &vsiz)) != NULL){
+    tmp = cburldecode(vbuf, NULL);
+    cbstrtolower(tmp);
+    enc = cburlencode(tmp, -1);
+    cbdatumcat(rbuf, enc, -1);
+    free(enc);
+    free(tmp);
+  } else {
+    cbdatumcat(rbuf, "localhost", -1);
+  }
+  if((vbuf = cbmapget(belems, "port", -1, &vsiz)) != NULL &&
+     (num = atoi(vbuf)) != port && num > 1){
+    sprintf(numbuf, ":%d", num);
+    cbdatumcat(rbuf, numbuf, -1);
+  }
+  if(!(path = cbmapget(telems, "path", -1, NULL))) path = "/";
+  if(path[0] == '\0' && (vbuf = cbmapget(belems, "path", -1, NULL)) != NULL) path = vbuf;
+  if(path[0] == '\0') path = "/";
+  bpaths = cblistopen();
+  if(path[0] != '/' && (vbuf = cbmapget(belems, "path", -1, &vsiz)) != NULL){
+    opaths = cbsplit(vbuf, vsiz, "/");
+  } else {
+    opaths = cbsplit("/", 1, "/");
+  }
+  free(cblistpop(opaths, NULL));
+  for(i = 0; i < CB_LISTNUM(opaths); i++){
+    vbuf = CB_LISTVAL2(opaths, i, &vsiz);
+    if(vsiz < 1 || !strcmp(vbuf, ".")) continue;
+    if(!strcmp(vbuf, "..")){
+      free(cblistpop(bpaths, NULL));
+    } else {
+      cblistpush(bpaths, vbuf, vsiz);
+    }
+  }
+  cblistclose(opaths);
+  opaths = cbsplit(path, -1, "/");
+  for(i = 0; i < CB_LISTNUM(opaths); i++){
+    vbuf = CB_LISTVAL2(opaths, i, &vsiz);
+    if(vsiz < 1 || !strcmp(vbuf, ".")) continue;
+    if(!strcmp(vbuf, "..")){
+      free(cblistpop(bpaths, NULL));
+    } else {
+      cblistpush(bpaths, vbuf, vsiz);
+    }
+  }
+  cblistclose(opaths);
+  for(i = 0; i < CB_LISTNUM(bpaths); i++){
+    vbuf = CB_LISTVAL(bpaths, i, NULL);
+    tmp = cburldecode(vbuf, NULL);
+    enc = cburlencode(tmp, -1);
+    cbdatumcat(rbuf, "/", 1);
+    cbdatumcat(rbuf, enc, -1);
+    free(enc);
+    free(tmp);
+  }
+  if(cbstrbwmatch(path, "/")) cbdatumcat(rbuf, "/", 1);
+  cblistclose(bpaths);
+  if((vbuf = cbmapget(telems, "query", -1, &vsiz)) != NULL){
+    cbdatumcat(rbuf, "?", 1);
+    qelems = cbsplit(vbuf, vsiz, "&;");
+    for(i = 0; i < CB_LISTNUM(qelems); i++){
+      vbuf = CB_LISTVAL(qelems, i, NULL);
+      if(i > 0) cbdatumcat(rbuf, "&", 1);
+      if((wp = strchr(vbuf, '=')) != NULL){
+        *wp = '\0';
+        tmp = cburldecode(vbuf, NULL);
+        enc = cburlencode(tmp, -1);
+        cbdatumcat(rbuf, enc, -1);
+        free(enc);
+        free(tmp);
+        cbdatumcat(rbuf, "=", 1);
+        wp++;
+        tmp = cburldecode(wp, NULL);
+        enc = cburlencode(tmp, -1);
+        cbdatumcat(rbuf, enc, -1);
+        free(enc);
+        free(tmp);
+      } else {
+        tmp = cburldecode(vbuf, NULL);
+        enc = cburlencode(tmp, -1);
+        cbdatumcat(rbuf, enc, -1);
+        free(enc);
+        free(tmp);
+      }
+    }
+    cblistclose(qelems);
+  }
+  if((vbuf = cbmapget(telems, "fragment", -1, &vsiz)) != NULL){
+    tmp = cburldecode(vbuf, NULL);
+    enc = cburlencode(tmp, -1);
+    cbdatumcat(rbuf, "#", 1);
+    cbdatumcat(rbuf, enc, -1);
+    free(enc);
+    free(tmp);
+  }
+  cbmapclose(belems);
   cbmapclose(telems);
   return cbdatumtomalloc(rbuf, NULL);
 }
@@ -1760,7 +1844,7 @@ char *cburlencode(const char *ptr, int size){
   for(i = 0; i < size; i++){
     c = ((unsigned char *)ptr)[i];
     if((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') ||
-       (c >= '0' && c <= '9') || (c != '\0' && strchr("_-.", c))){
+       (c >= '0' && c <= '9') || (c != '\0' && strchr("_-.!~*'()", c))){
       *(wp++) = c;
     } else {
       wp += sprintf(wp, "%%%02X", c);
@@ -1773,7 +1857,6 @@ char *cburlencode(const char *ptr, int size){
 
 /* Decode a string encoded with URL encoding. */
 char *cburldecode(const char *str, int *sp){
-  const char *hex = "1234567890abcdefABCDEF";
   char *buf, *wp;
   unsigned char c;
   buf = cbmemdup(str, -1);
@@ -1781,7 +1864,10 @@ char *cburldecode(const char *str, int *sp){
   while(*str != '\0'){
     if(*str == '%'){
       str++;
-      if(strchr(hex, *str) && strchr(hex, *(str + 1))){
+      if(((str[0] >= '0' && str[0] <= '9') || (str[0] >= 'A' && str[0] <= 'F') ||
+          (str[0] >= 'a' && str[0] <= 'f')) &&
+         ((str[1] >= '0' && str[1] <= '9') || (str[1] >= 'A' && str[1] <= 'F') ||
+          (str[1] >= 'a' && str[1] <= 'f'))){
         c = *str;
         if(c >= 'A' && c <= 'Z') c += 'a' - 'A';
         if(c >= 'a' && c <= 'z'){
@@ -2711,7 +2797,11 @@ time_t cbstrmktime(const char *str){
   time_t t;
   struct tm ts;
   assert(str);
-  if(cbstrfwimatch(str, "0x")) return (time_t)strtol(str + 2, NULL, 16);
+  while(*str > '\0' && *str <= ' '){
+    str++;
+  }
+  if(*str == '\0') return -1;
+  if(str[0] == '0' && (str[1] == 'x' || str[1] == 'X')) return (time_t)strtol(str + 2, NULL, 16);
   memset(&ts, 0, sizeof(struct tm));
   ts.tm_year = 70;
   ts.tm_mon = 0;
@@ -2722,7 +2812,16 @@ time_t cbstrmktime(const char *str){
   ts.tm_isdst = 0;
   len = strlen(str);
   t = (time_t)strtol(str, &pv, 10);
-  if(*pv == '\0') return t;
+  if(*pv >= '\0' && *pv <= ' '){
+    while(*pv > '\0' && *pv <= ' '){
+      pv++;
+    }
+    if(*pv == '\0') return t;
+  }
+  if((pv[0] == 's' || pv[0] == 'S') && pv[1] >= '\0' && pv[1] <= ' ') return t;
+  if((pv[0] == 'm' || pv[0] == 'M') && pv[1] >= '\0' && pv[1] <= ' ') return t * 60;
+  if((pv[0] == 'h' || pv[0] == 'H') && pv[1] >= '\0' && pv[1] <= ' ') return t * 60 * 60;
+  if((pv[0] == 'd' || pv[0] == 'D') && pv[1] >= '\0' && pv[1] <= ' ') return t * 60 * 60 * 24;
   if(len > 4 && str[4] == '-'){
     ts.tm_year = atoi(str) - 1900;
     if((pv = strchr(str, '-')) != NULL && pv - str == 4){
@@ -2742,9 +2841,9 @@ time_t cbstrmktime(const char *str){
             rp = pv + 1;
             ts.tm_sec = atoi(rp);
           }
-          if((pv = strchr(rp, '.')) != NULL && pv - str == 19) rp = pv + 1;
+          if((pv = strchr(rp, '.')) != NULL && pv - str >= 19) rp = pv + 1;
           strtol(rp, &pv, 10);
-          if((*pv == '+' || *pv == '-') && strlen(pv) == 6 && pv[3] == ':')
+          if((*pv == '+' || *pv == '-') && strlen(pv) >= 6 && pv[3] == ':')
             ts.tm_sec -= (atoi(pv + 1) * 3600 + atoi(pv + 4) * 60) * (pv[0] == '+' ? 1 : -1);
         }
       }
@@ -2771,9 +2870,9 @@ time_t cbstrmktime(const char *str){
             rp = pv + 1;
             ts.tm_sec = atoi(rp);
           }
-          if((pv = strchr(rp, '.')) != NULL && pv - str == 19) rp = pv + 1;
+          if((pv = strchr(rp, '.')) != NULL && pv - str >= 19) rp = pv + 1;
           strtol(rp, &pv, 10);
-          if((*pv == '+' || *pv == '-') && strlen(pv) == 6 && pv[3] == ':')
+          if((*pv == '+' || *pv == '-') && strlen(pv) >= 6 && pv[3] == ':')
             ts.tm_sec -= (atoi(pv + 1) * 3600 + atoi(pv + 4) * 60) * (pv[0] == '+' ? 1 : -1);
         }
       }
@@ -2835,7 +2934,7 @@ time_t cbstrmktime(const char *str){
       ts.tm_hour = atoi(crp + 0);
       ts.tm_min = atoi(crp + 3);
       ts.tm_sec = atoi(crp + 6);
-      if(clen == 14 && crp[8] == ' ' && (crp[9] == '+' || crp[9] == '-')){
+      if(clen >= 14 && crp[8] == ' ' && (crp[9] == '+' || crp[9] == '-')){
         ts.tm_sec -= ((crp[10] - '0') * 36000 + (crp[11] - '0') * 3600 +
                       (crp[12] - '0') * 600 + (crp[13] - '0') * 60) * (crp[9] == '+' ? 1 : -1);
       } else if(clen > 9){
@@ -2983,19 +3082,19 @@ void cbmapputvbuf(CBMAP *map, const char *kbuf, int ksiz, char *vbuf, int vsiz){
   hash = cbsecondhash(kbuf, ksiz);
   while(datum){
     if(hash > datum->hash){
-      entp = (CBMAPDATUM **)&(datum->left);
-      datum = (CBMAPDATUM *)datum->left;
+      entp = &(datum->left);
+      datum = datum->left;
     } else if(hash < datum->hash){
-      entp = (CBMAPDATUM **)&(datum->right);
-      datum = (CBMAPDATUM *)datum->right;
+      entp = &(datum->right);
+      datum = datum->right;
     } else {
       kcmp = cbkeycmp(kbuf, ksiz, datum->kbuf, datum->ksiz);
       if(kcmp < 0){
-        entp = (CBMAPDATUM **)&(datum->left);
-        datum = (CBMAPDATUM *)datum->left;
+        entp = &(datum->left);
+        datum = datum->left;
       } else if(kcmp > 0){
-        entp = (CBMAPDATUM **)&(datum->right);
-        datum = (CBMAPDATUM *)datum->right;
+        entp = &(datum->right);
+        datum = datum->right;
       } else {
         free(datum->vbuf);
         datum->vbuf = vbuf;
@@ -3012,11 +3111,11 @@ void cbmapputvbuf(CBMAP *map, const char *kbuf, int ksiz, char *vbuf, int vsiz){
   datum->hash = hash;
   datum->left = NULL;
   datum->right = NULL;
-  datum->prev = (char *)map->last;
+  datum->prev = map->last;
   datum->next = NULL;
   *entp = datum;
   if(!map->first) map->first = datum;
-  if(map->last) map->last->next = (char *)datum;
+  if(map->last) map->last->next = datum;
   map->last = datum;
   map->rnum++;
 }
