@@ -398,10 +398,9 @@ char *crname(CURIA *curia){
 int crfsiz(CURIA *curia){
   int i, sum, rv;
   assert(curia);
-  sum = 0;
+  if((sum = dpfsiz(curia->attr)) == -1) return -1;
   for(i = 0; i < curia->dnum; i++){
-    rv = dpfsiz(curia->depots[i]);
-    if(rv == -1) return -1;
+    if((rv = dpfsiz(curia->depots[i])) == -1) return -1;
     sum += rv;
   }
   return sum;
@@ -414,9 +413,9 @@ double crfsizd(CURIA *curia){
   int i, rv;
   assert(curia);
   sum = 0.0;
+  if((sum = dpfsiz(curia->attr)) < 0) return -1.0;
   for(i = 0; i < curia->dnum; i++){
-    rv = dpfsiz(curia->depots[i]);
-    if(rv == -1) return -1.0;
+    if((rv = dpfsiz(curia->depots[i])) == -1) return -1.0;
     sum += rv;
   }
   return sum;
