@@ -607,6 +607,16 @@ int domisc(const char* name){
       if(zvilla.vnum(buf, len) != 2) throw Villa_error();
     }
     cout << "ok" << endl;
+    // cursor operations
+    cout << "Cursor operations ... ";
+    zvilla.curjump("00000010", -1, Villa::JFORWARD);
+    zvilla.curput("CURRENT", -1, Villa::CPCURRENT);
+    zvilla.curput("BEFORE", -1, Villa::CPBEFORE);
+    zvilla.curput("AFTER", -1, Villa::CPAFTER);
+    zvilla.curnext();
+    zvilla.curout();
+    zvilla.curput("OMIT", -1);
+    cout << "ok" << endl;
     // close the database
     cout << "Closing the database ... ";
     zvilla.close();
