@@ -1,6 +1,6 @@
 #=================================================================================================
 # Ruby API of Depot, the basic API of QDBM
-#                                                       Copyright (C) 2000-2005 Mikio Hirabayashi
+#                                                       Copyright (C) 2000-2006 Mikio Hirabayashi
 # This file is part of QDBM, Quick Database Manager.
 # QDBM is free software; you can redistribute it and/or modify it under the terms of the GNU
 # Lesser General Public License as published by the Free Software Foundation; either version
@@ -254,7 +254,7 @@ class Depot
   # `align' specifies the basic size of alignment.  If it is omitted, alignment is cleared.
   # The return value is always `true'.
   # An exception of `Depot::EANY' or its sub classes is thrown if an error occurs.
-  # If alignment is set to a database, the efficiency of overwriting values are improved.
+  # If alignment is set to a database, the efficiency of overwriting values is improved.
   # The size of alignment is suggested to be average size of the values of the records to be
   # stored.  If alignment is positive, padding whose size is multiple number of the alignment
   # is placed.  If alignment is negative, as `vsiz' is the size of a value, the size of padding
@@ -265,6 +265,19 @@ class Depot
     mod_setalign(@index, align)
   end
   ##
+  # bool = depot.setfbpsiz(size);
+  # Method: Set the size of the free block pool.
+  # `size' specifies the size of the free block pool.  If it is undef, the free block pool is not
+  # used.
+  # The return value is always `true'.
+  # An exception of `Depot::EANY' or its sub classes is thrown if an error occurs.
+  # The default size of the free block pool is 16.  If the size is greater, the space efficiency
+  # of overwriting values is improved with the time efficiency sacrificed.
+  #
+  def setfbpsiz(size = 0)
+    mod_setfbpsiz(@index, size)
+  end
+  ##
   # bool = depot.sync()
   # Method: Synchronize updating contents with the file and the device.
   # The return value is always `true'.
