@@ -24,7 +24,6 @@
 #define VL_VNUMBUFSIZ  8                 /* size of a buffer for variable length number */
 #define VL_NUMBUFSIZ   32                /* size of a buffer for a number */
 #define VL_PAGEBUFSIZ  32768             /* size of a buffer to read each page */
-#define VL_LEVELMAX    64                /* max level of B+ tree */
 #define VL_DEFLRECMAX  49                /* default number of records in each leaf */
 #define VL_DEFNIDXMAX  192               /* default number of indexes in each node */
 #define VL_DEFLCNUM    1024              /* default number of leaf cache */
@@ -62,6 +61,7 @@ static VLLEAF *vlleafnew(VILLA *villa, int prev, int next);
 static int vlleafcacheout(VILLA *villa, int id);
 static int vlleafsave(VILLA *villa, VLLEAF *leaf);
 static VLLEAF *vlleafload(VILLA *villa, int id);
+static VLLEAF *vlgethistleaf(VILLA *villa, const char *kbuf, int ksiz);
 static int vlleafaddrec(VILLA *villa, VLLEAF *leaf, int dmode,
                         const char *kbuf, int ksiz, const char *vbuf, int vsiz);
 static VLLEAF *vlleafdivide(VILLA *villa, VLLEAF *leaf);
@@ -71,7 +71,7 @@ static int vlnodesave(VILLA *villa, VLNODE *node);
 static VLNODE *vlnodeload(VILLA *villa, int id);
 static void vlnodeaddidx(VILLA *villa, VLNODE *node, int order,
                          int pid, const char *kbuf, int ksiz);
-static int vlsearchleaf(VILLA *villa, const char *kbuf, int ksiz, int *hist, int *hnp);
+static int vlsearchleaf(VILLA *villa, const char *kbuf, int ksiz);
 static int vlcacheadjust(VILLA *villa);
 static VLREC *vlrecsearch(VILLA *villa, VLLEAF *leaf, const char *kbuf, int ksiz, int *ip);
 
@@ -165,6 +165,9 @@ VILLA *vlopen(const char *name, int omode, VLCFUNC cmp){
   villa->rnum = rnum;
   villa->leafc = cbmapopen();
   villa->nodec = cbmapopen();
+  villa->hnum = 0;
+  villa->hleaf = -1;
+  villa->lleaf = -1;
   villa->curleaf = -1;
   villa->curknum = -1;
   villa->curvnum = -1;
@@ -232,8 +235,7 @@ int vlput(VILLA *villa, const char *kbuf, int ksiz, const char *vbuf, int vsiz,
   VLNODE *node, *newnode;
   VLIDX *idxp;
   CBDATUM *key;
-  int hist[VL_LEVELMAX];
-  int i, hnum, pid, heir, parent, mid;
+  int i, pid, heir, parent, mid;
   assert(villa && kbuf && vbuf);
   villa->curleaf = -1;
   villa->curknum = -1;
@@ -244,8 +246,10 @@ int vlput(VILLA *villa, const char *kbuf, int ksiz, const char *vbuf, int vsiz,
   }
   if(ksiz < 0) ksiz = strlen(kbuf);
   if(vsiz < 0) vsiz = strlen(vbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, hist, &hnum)) == -1) return FALSE;
-  if(!(leaf = vlleafload(villa, pid))) return FALSE;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return FALSE;
+    if(!(leaf = vlleafload(villa, pid))) return FALSE;
+  }
   if(!vlleafaddrec(villa, leaf, dmode, kbuf, ksiz, vbuf, vsiz)){
     dpecodeset(DP_EKEEP, __FILE__, __LINE__);
     return FALSE;
@@ -258,14 +262,14 @@ int vlput(VILLA *villa, const char *kbuf, int ksiz, const char *vbuf, int vsiz,
     key = ((VLREC *)CB_LISTVAL(newleaf->recs, 0, NULL))->key;
     key = cbdatumopen(CB_DATUMPTR(key), CB_DATUMSIZE(key));
     while(TRUE){
-      if(hnum < 1){
+      if(villa->hnum < 1){
         node = vlnodenew(villa, heir);
         vlnodeaddidx(villa, node, TRUE, pid, CB_DATUMPTR(key), CB_DATUMSIZE(key));
         villa->root = node->id;
         cbdatumclose(key);
         break;
       }
-      parent = hist[--hnum];
+      parent = villa->hist[--villa->hnum];
       if(!(node = vlnodeload(villa, parent))){
         cbdatumclose(key);
         return FALSE;
@@ -312,8 +316,10 @@ int vlout(VILLA *villa, const char *kbuf, int ksiz){
     return FALSE;
   }
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1) return FALSE;
-  if(!(leaf = vlleafload(villa, pid))) return FALSE;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return FALSE;
+    if(!(leaf = vlleafload(villa, pid))) return FALSE;
+  }
   if(!(recp = vlrecsearch(villa, leaf, kbuf, ksiz, &ri))){
     dpecodeset(DP_ENOITEM, __FILE__, __LINE__);
     return FALSE;
@@ -346,8 +352,10 @@ char *vlget(VILLA *villa, const char *kbuf, int ksiz, int *sp){
   int pid;
   assert(villa && kbuf);
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1) return NULL;
-  if(!(leaf = vlleafload(villa, pid))) return NULL;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return NULL;
+    if(!(leaf = vlleafload(villa, pid))) return NULL;
+  }
   if(!(recp = vlrecsearch(villa, leaf, kbuf, ksiz, NULL))){
     dpecodeset(DP_ENOITEM, __FILE__, __LINE__);
     return NULL;
@@ -365,8 +373,10 @@ int vlvsiz(VILLA *villa, const char *kbuf, int ksiz){
   int pid;
   assert(villa && kbuf);
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1) return -1;
-  if(!(leaf = vlleafload(villa, pid))) return -1;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return -1;
+    if(!(leaf = vlleafload(villa, pid))) return -1;
+  }
   if(!(recp = vlrecsearch(villa, leaf, kbuf, ksiz, NULL))){
     dpecodeset(DP_ENOITEM, __FILE__, __LINE__);
     return -1;
@@ -383,8 +393,10 @@ int vlvnum(VILLA *villa, const char *kbuf, int ksiz){
   int pid;
   assert(villa && kbuf);
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1) return 0;
-  if(!(leaf = vlleafload(villa, pid))) return 0;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return 0;
+    if(!(leaf = vlleafload(villa, pid))) return 0;
+  }
   if(!(recp = vlrecsearch(villa, leaf, kbuf, ksiz, NULL))){
     dpecodeset(DP_ENOITEM, __FILE__, __LINE__);
     return 0;
@@ -442,8 +454,10 @@ CBLIST *vlgetlist(VILLA *villa, const char *kbuf, int ksiz){
   const char *vbuf;
   assert(villa && kbuf);
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1) return NULL;
-  if(!(leaf = vlleafload(villa, pid))) return NULL;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return NULL;
+    if(!(leaf = vlleafload(villa, pid))) return NULL;
+  }
   if(!(recp = vlrecsearch(villa, leaf, kbuf, ksiz, NULL))){
     dpecodeset(DP_ENOITEM, __FILE__, __LINE__);
     return NULL;
@@ -473,8 +487,10 @@ char *vlgetcat(VILLA *villa, const char *kbuf, int ksiz, int *sp){
   const char *vbuf;
   assert(villa && kbuf);
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1) return NULL;
-  if(!(leaf = vlleafload(villa, pid))) return NULL;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return NULL;
+    if(!(leaf = vlleafload(villa, pid))) return NULL;
+  }
   if(!(recp = vlrecsearch(villa, leaf, kbuf, ksiz, NULL))){
     dpecodeset(DP_ENOITEM, __FILE__, __LINE__);
     return NULL;
@@ -665,7 +681,7 @@ int vlcurjump(VILLA *villa, const char *kbuf, int ksiz, int jmode){
   int pid, index;
   assert(villa && kbuf);
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1){
+  if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1){
     villa->curleaf = -1;
     return FALSE;
   }
@@ -1412,8 +1428,10 @@ const char *vlgetcache(VILLA *villa, const char *kbuf, int ksiz, int *sp){
   int pid;
   assert(villa && kbuf);
   if(ksiz < 0) ksiz = strlen(kbuf);
-  if((pid = vlsearchleaf(villa, kbuf, ksiz, NULL, NULL)) == -1) return NULL;
-  if(!(leaf = vlleafload(villa, pid))) return NULL;
+  if(villa->hleaf < VL_LEAFIDMIN || !(leaf = vlgethistleaf(villa, kbuf, ksiz))){
+    if((pid = vlsearchleaf(villa, kbuf, ksiz)) == -1) return NULL;
+    if(!(leaf = vlleafload(villa, pid))) return NULL;
+  }
   if(!(recp = vlrecsearch(villa, leaf, kbuf, ksiz, NULL))){
     dpecodeset(DP_ENOITEM, __FILE__, __LINE__);
     return NULL;
@@ -2101,6 +2119,29 @@ static VLLEAF *vlleafload(VILLA *villa, int id){
 }
 
 
+/* Load the historical leaf from the database.
+   `villa' specifies a database handle.
+   `kbuf' specifies the pointer to the region of a key.
+   `ksiz' specifies the size of the region of the key.
+   If successful, the return value is the pointer to the leaf, else, it is `NULL'. */
+static VLLEAF *vlgethistleaf(VILLA *villa, const char *kbuf, int ksiz){
+  VLLEAF *leaf;
+  VLREC *recp;
+  int ln, rv;
+  assert(villa);
+  if(!(leaf = vlleafload(villa, villa->hleaf))) return NULL;
+  if((ln = CB_LISTNUM(leaf->recs)) < 2) return NULL;
+  recp = (VLREC *)CB_LISTVAL(leaf->recs, 0, NULL);
+  rv = villa->cmp(kbuf, ksiz, CB_DATUMPTR(recp->key), CB_DATUMSIZE(recp->key));
+  if(rv == 0) return leaf;
+  if(rv < 0) return NULL;
+  recp = (VLREC *)CB_LISTVAL(leaf->recs, ln - 1, NULL);
+  rv = villa->cmp(kbuf, ksiz, CB_DATUMPTR(recp->key), CB_DATUMSIZE(recp->key));
+  if(rv <= 0 || leaf->next < VL_LEAFIDMIN) return leaf;
+  return NULL;
+}
+
+
 /* Add a record to a leaf.
    `villa' specifies a database handle.
    `leaf' specifies a leaf handle.
@@ -2197,6 +2238,7 @@ static VLLEAF *vlleafdivide(VILLA *villa, VLLEAF *leaf){
   VLREC *recp;
   int i, mid, ln;
   assert(villa && leaf);
+  villa->hleaf = -1;
   mid = CB_LISTNUM(leaf->recs) / 2;
   recp = (VLREC *)CB_LISTVAL(leaf->recs, mid, NULL);
   newleaf = vlleafnew(villa, leaf->id, leaf->next);
@@ -2407,25 +2449,22 @@ static void vlnodeaddidx(VILLA *villa, VLNODE *node, int order,
    `villa' specifies a database handle.
    `kbuf' specifies the pointer to the region of a key.
    `ksiz' specifies the size of the region of the key.
-   `hist' specifies an array history of visited nodes.  If `NULL', it is not used.
-   `hnp' specifies the pointer to a variable to which the number of elements of the history
-   assigned.  If `NULL', it is not used.
    The return value is the ID number of the leaf, or -1 on failure. */
-static int vlsearchleaf(VILLA *villa, const char *kbuf, int ksiz, int *hist, int *hnp){
+static int vlsearchleaf(VILLA *villa, const char *kbuf, int ksiz){
   VLNODE *node;
   VLIDX *idxp;
-  int i, pid, level, rv, left, right, ln;
+  int i, pid, rv, left, right, ln;
   assert(villa && kbuf && ksiz >= 0);
   pid = villa->root;
   idxp = NULL;
-  level = 0;
+  villa->hnum = 0;
+  villa->hleaf = -1;
   while(pid >= VL_NODEIDMIN){
     if(!(node = vlnodeload(villa, pid)) || (ln = CB_LISTNUM(node->idxs)) < 1){
       dpecodeset(DP_EBROKEN, __FILE__, __LINE__);
-      if(hnp) *hnp = level;
       return -1;
     }
-    if(hist) hist[level++] = node->id;
+    villa->hist[villa->hnum++] = node->id;
     left = 1;
     right = ln;
     i = (left + right) / 2;
@@ -2457,7 +2496,8 @@ static int vlsearchleaf(VILLA *villa, const char *kbuf, int ksiz, int *hist, int
     }
     if(i >= ln) pid = idxp->pid;
   }
-  if(hnp) *hnp = level;
+  if(villa->lleaf == pid) villa->hleaf = pid;
+  villa->lleaf = pid;
   return pid;
 }
 
