@@ -41,6 +41,8 @@ extern "C" {
  *************************************************************************************************/
 
 
+#define VL_LEVELMAX    64                /* max level of B+ tree */
+
 typedef struct {                         /* type of structure for a record */
   CBDATUM *key;                          /* datum of the key */
   CBDATUM *first;                        /* datum of the first value */
@@ -92,6 +94,10 @@ typedef struct {                         /* type of structure for a database han
   int rnum;                              /* number of records */
   CBMAP *leafc;                          /* cache for leaves */
   CBMAP *nodec;                          /* cache for nodes */
+  int hist[VL_LEVELMAX];                 /* array history of visited nodes */
+  int hnum;                              /* number of elements of the history */
+  int hleaf;                             /* ID number of the leaf referred by the history */
+  int lleaf;                             /* ID number of the last visited leaf */
   int curleaf;                           /* ID number of the leaf where the cursor is */
   int curknum;                           /* index of the key where the cursor is */
   int curvnum;                           /* index of the value where the cursor is */
